################################################################################
# Box Class                                                                    #
#                                                                              #
"""Create simulation box object."""
################################################################################


class Box:
    """This class creates a hypothetical simulation box, to which molecules are
    added. Furthermore, all properties like temperature and specific job
    properties can be adjusted separate from the global simulation values.

    Parameters
    ----------
    name : string, optional
        Simulation name
    label : None, string, optional
        Job identification label
    """
    def __init__(self, name="box", label=None):
        # Initialize
        self._sim_dict = {}
        self._sim_dict["name"] = [name, name if label is None else label]

        self._sim_dict["struct"] = {}
        self._sim_dict["mols"] = {}
        self._sim_dict["topol"] = {"master":[], "top":[], "itp":[]}


    ##################
    # Public Methods #
    ##################
    def add_box(self, link):
        """Set simulation box structure link.

        Parameters
        ----------
        link : string
            Simulation box structure link
        """
        self._sim_dict["struct"]["BOX"] = link

    def add_pore(self, link):
        """Add pore object to structure folder.

        Parameters
        ----------
        link : string
            Pore object link
        """
        self._sim_dict["struct"]["PORE"] = link

    def add_mol(self, short, link, inp, num_atoms="gro", auto_dens=None, mass = None,  section="both",area= [], box = [], kwargs_gmx = {}):
        """Add a molecule to the simulation box. A unique short name and a
        structure-file link have to be given.

        Furthermore, the input is either a specific number of molecules or the
        input **fill** in order to continuously fill the simulation box with the
        specified molecule. This can be done automatically by defining the
        reference density to fill up to.

        The number atoms can either be specified numerically or extracted, which
        currently only works for the GROMACS file-type.

        Parameters
        ----------
        short : string
            Molecule short name as given in the topology
        link : string
            Molecule structure file link
        inp : integer, string
            Define number of molecules or **fill** for filling a pore or box
        num_atoms : integer, string, optional
            Number of atoms in the specified molecule or filetype for extraction
        auto_dens : float, None, optional
            Density in :math:`\\frac{\\text{kg}}{\\text{m}^3}`
        mass : float, None, optional
            molecular mass (:math:`\\frac{\\text{g}}{\\text{mol}}`) of the molecule to get an estimate of how many molecules you need to put in the box as the initial value for auto_dens.
        section : string, "both", optional
            'Both' the molecules in the pore and in the reservoir can be set. 
            If you want the molecules only in the pore, choose 'pore'. If you want them only in the reservoir, choose 'res'. 
            For a slit pore you can use wall to put the molecules as a layer on the SiO2 surface.
            Chose "box" for a box system.
        area : list, [], optional
            To set up a 2-phase box system without a pore, specify the areas of the section [a,b] where the molecules should be inserted using Gromacs.
        box : list, [], optional
            To set up a 2-phase system, specify the dimensions [x, y, z] of the simulation box.
        kwargs_gmx : dictonary, {}, optional
            dictonary to set up options for the gromacs "insert-molecules" command (adjust for example -try and -scale)
        """
        # Process input
        if not isinstance(inp, int) and not isinstance(inp, str):
            print("Wrong input for the molecule number ...")
            return

        if not isinstance(num_atoms, int) and not num_atoms in ["gro"]:
            print("Wrong input for the number of atoms ...")
            return

        if auto_dens is not None and not (isinstance(auto_dens, float) or isinstance(auto_dens, int)):
            print("Wrong input for the auto density option ...")
            return

        # Extract atom number
        if num_atoms == "gro":
            counter = 1
            with open(link, "r") as file_in:
                for line in file_in:
                    if counter == 2:
                        num_atoms = int(line)
                        break
                    else:
                        counter += 1

        # If Fill
        if inp=="fill" and not mass:
            print("If you use auto_dens you have to specifiy the molar mass of the molecule")
            return

        # Add to global list
        self._sim_dict["struct"][short] = link
        self._sim_dict["mols"][short] = [inp, num_atoms, auto_dens, mass,  section,area, box, kwargs_gmx]

    def add_struct(self, ident, link):
        """Add file link to the structure dictionary.

        Parameters
        ----------
        ident : string
            File identifier
        link : string
            File link
        """
        self._sim_dict["struct"][ident] = link

    def add_topol(self, link, top_type="itp"):
        """Add simulation topology files to simulation dictionary. For topology
        files, three different types exist with different handlings

        * **master** - Master topology file will be named topol.top with an additional backup
        * **top** - Topology file will be copied
        * **itp** - Topology file will be filtered for an itp file

        Parameters
        ----------
        link : string, list
            Link or a list of those containing a topology file link
        top_type : string
            Topology type
        """
        links = link if isinstance(link, list) else [link]
        for link in links:
            self._sim_dict["topol"][top_type].append(link)

    def add_charge_si(self, charge):
        """
        Add silicon charge for grid-atoms

        Parameters
        ----------
        charge : float
            Silicon charge for pore grid molecules
        """
        self._sim_dict["topol"]["charge"] = charge


    ##################
    # Setter Methods #
    ##################
    def set_sim_dict(self, sim_dict):
        """Set the simulation dictionary.

        Parameters
        ----------
        sim_dict : dictionary
            Simulation dictionary
        """
        self._sim_dict = sim_dict

    def set_name(self, name):
        """Set the simulation box name.

        Parameters
        ----------
        name : string
            Simulation box name
        """
        self._sim_dict["name"][0] = name

    def set_label(self, label):
        """Set the simulation box label, for job identification.

        Parameters
        ----------
        label : string
            Label name
        """
        self._sim_dict["name"][1] = label

    def set_job(self, job):
        """Set the dictionary containing all simulation information.

        Parameters
        ----------
        job : dictionary
            Information dictionary

        Examples
        --------
        Following dictionary structure may be used as an input

        .. code-block:: python

            {"min": {"file": "data/simulation/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
             "nvt": {"file": "data/simulation/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
             "run": {"file": "data/simulation/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}
        """
        self._sim_dict["job"] = job

    def set_param(self, param):
        """Set the dictionary containing all systems parameters.

        Parameters
        ----------
        param : dictionary
            Information dictionary

        Examples
        --------
        Following dictionary structure may be used as an input

        .. code-block:: python

            {"min": {"file": "data/simulation/pore_min.mdp"},
             "nvt": {"file": "data/simulation/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
             "run": {"file": "data/simulation/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}
        """
        self._sim_dict["param"] = param


    ##################
    # Getter Methods #
    ##################
    def get_sim_dict(self):
        """Return the simulation dictionary.

        Returns
        -------
        sim_dict : dictionary
            Simulation dictionary
        """
        return self._sim_dict

    def get_name(self):
        """Return the simulation box name.

        Returns
        -------
        name : string
            Simulation box name
        """
        return self._sim_dict["name"][0]

    def get_label(self):
        """Return the simulation label.

        Returns
        -------
        label : string
            Simulation label
        """
        return self._sim_dict["name"][1]

    def get_job(self):
        """Return the dictionary containing all simulation information.

        Returns
        -------
        job : dictionary
            Simulation dictionary
        """
        return self._sim_dict["job"] if "job" in self._sim_dict.keys() else None

    def get_param(self):
        """Return the dictionary containing all systems parameters.

        Returns
        -------
        param : dictionary
            Parameter dictionary
        """
        return self._sim_dict["param"] if "param" in self._sim_dict.keys() else None

    def get_mols(self):
        """Return list of molecule numbers to be filled in the box.

        Returns
        -------
        mols : dictionary
            List of molecules
        """
        return self._sim_dict["mols"]

    def get_struct(self):
        """Return list of structure files.

        Returns
        -------
        struct : dictionary
            Molecule structure file links
        """
        return self._sim_dict["struct"]

    def get_topol(self):
        """Return list of topology files.

        Returns
        -------
        topol : list
            List of topology file links
        """
        return self._sim_dict["topol"]



import os
import porems as pms
import poresim as ps
import poreana as pa


if __name__ == "__main__":
    # Todo
    print("Finish fill scripts ...")
    print("Finish ana.sh file ...")
    print("Add following script to running shell")
    # cd ../ana
    # sh ana.sh
    # python ana.py

    # Load molecule
    mol_dict = {}
    {% for mol in mols -%}
    mol_dict["{{mol.name }}"] = pms.Molecule("molecule", "{{mol.name }}", inp="{{mol.link }}")
    {% endfor %}


    ## Silanol
    sioh = pms.Molecule("sioh", "SL")
    sioh.add("Si", [0, 0, 0], name="Si1")
    sioh.add("O", 0, r=1.3, name="O1")
    sioh.add("H", 1, r=1.0, name="H1")

    # Load pore
    pore = pa.utils.load("../_gro/pore.yml")

    # Set analysis
    ana_list = {}
    {% for mol in mols -%}
    ana_list["{{mol.name }}"] = {"traj": "traj_{{mol.name }}.xtc", "dens": True, "dens_box": True, "diff": False, "mc_trans": False, "mc": False, "mol": mol_dict["{{mol.name }}"], "atoms": []}
    {% endfor %}

    # ana_list["sioh"] = {"traj": "traj_sioh.xtc", "dens": True, "dens_box": True, "diff": False, "mc_trans": False, "mc": False, "mol": sioh, "atoms": ["O1"]}

    # Run analysis
    for ana_name, ana_props in ana_list.items():
        if ana_props["dens_box"]:
            sample = pa.Sample(pore["system"]["dimensions"], ana_props["traj"], ana_props["mol"], ana_props["atoms"], [1 for x in ana_props["atoms"]])
            sample.init_density("dens_"+ana_name+"_box.obj", remove_pore_from_res=True)
            sample.sample(is_parallel=True)

        if ana_props["dens"] or ana_props["diff"]:
            sample = pa.Sample("../_gro/pore.yml", ana_props["traj"], ana_props["mol"], ana_props["atoms"], [1 for x in ana_props["atoms"]])
            if ana_props["dens"]:
                sample.init_density("dens_"+ana_name+".obj", remove_pore_from_res=False)
            if ana_props["diff"]:
                sample.init_diffusion_bin("diff_"+ana_name+".obj", bin_num=35)
            sample.sample(is_parallel=True)

        if ana_props["mc_trans"]:
            sample = pa.Sample("../_gro/pore.yml", ana_props["traj"], ana_props["mol"], ana_props["atoms"], [1 for x in ana_props["atoms"]])
            sample.init_diffusion_mc("diff_"+ana_name+"_trans.obj", [1, 2, 5, 10, 20, 30, 40, 50, 60, 70])
            sample.sample(is_parallel=True)

        if ana_props["mc"]:
            model = pa.CosineModel("diff_"+ana_name+"_trans.obj", 6, 10)
            pa.MC().run(model, "diff_"+ana_name+"_mc_cos.obj", nmc_eq=1000000, nmc=1000000)

    {% if fill %}
    # Automation
    is_auto = True
    if is_auto:
        # Calculate density - area is given in bins
        dens = {}
        {% for mol in mols2 -%}
        dens["{{mol.name }}"] = pa.density.bins("dens_{{mol.name }}.obj", target_dens={{mol.target_dens }}, area=[[0, 100], [0, 100]])
        {% endfor %}
        # Fill and rerun
        num_diff = {}
        {% for mol in mols2 -%}
        num_diff["{{mol.name }}"] = dens["{{mol.name }}"]["diff"]       
        with open("../_gro/" + "position_{}.dat".format("{{mol.name }}"), "w") as file_out:
            for i in range(int(num_diff["{{mol.name }}"]/2)):
                out_string = str(pore["system"]["dimensions"][0]/2) + " "
                out_string += str(pore["system"]["dimensions"][1]/2) + " "
                out_string += str(pore["system"]["reservoir"]/2) + "\n"
                file_out.write(out_string)
            for i in range(int(num_diff["{{mol.name }}"]/2)):
                out_string = str(pore["system"]["dimensions"][0]/2) + " "
                out_string += str(pore["system"]["dimensions"][1]/2) + " "
                out_string += str(pore["system"]["dimensions"][2] -pore["system"]["reservoir"]/2) + "\n"
                file_out.write(out_string)
            file_out.close()
        {% endfor %}
        if (all(i<10 for i in num_diff.values()))==False:
            ps.utils.copy("../_fill/fillBackup.sh", "../_fill/fill.sh")
            {% for mol in mols2 -%}
            if num_diff["{{mol.name }}"] > 10:
                ps.utils.replace("../_fill/fill.sh", "FILLDENS_{{mol.name }}", str(int(num_diff["{{mol.name }}"])))
            else:
                ps.utils.replace("../_fill/fill.sh", "FILLDENS_{{mol.name }}", str(int(0)))
            {% endfor %}
            os.system("cd ../_fill;sh fill.sh;cd ../min;{{submit }}")

    {% else %}
    # Nothing to fill up
    {% endif %}   



################################################################################
# Parameter Class                                                              #
#                                                                              #
"""Process parameter files."""
################################################################################


import poresim.utils as utils


class Parameter:
    """This class generates parameter files by replacing the given parameter in
    prepared files.

    Parameters
    ----------
    link : string
        Simulation box link
    param : dictionary
        Parameter dictionary
    """
    def __init__(self, link, param):
        # Initialize
        self._link = link+"_mdp/"
        self._param = param

        # Create folder
        utils.mkdirp(self._link)

    ##################
    # Public Methods #
    ##################
    def generate_files(self):
        """Create a parameter file for simulation based on the given parameter
        dictionary of the simulation box.

        The information for each simulation step should contain a file link, and
        optionally a list of parameters, to be replaced in the file. Parameter
        have the syntax of replacing string followed by the value.

        Examples
        --------
        >>> TEMPERATURE_VAL: 298
        """
        # Run through mdp files
        for step in self._param:
            file_name = step+".mdp"
            utils.copy(self._param[step]["file"], self._link+file_name)

            if "param" in self._param[step]:
                for inp in self._param[step]["param"]:
                    utils.replace(self._link+file_name, inp, str(self._param[step]["param"][inp]))



################################################################################
# Topology Class                                                               #
#                                                                              #
"""Process topology files."""
################################################################################


import poresim.utils as utils


class Topology:
    """This class reads and extract all information from topology files of the
    molecules specified in the simulation box. The aim was creating an
    independent class for editing and updating the topologies if necessary.

    Further editing functions should be easy to add.

    Parameters
    ----------
    link : string
        Simulation box link
    topol : dictionary
        Topology dictionary
    """
    def __init__(self, link, topol):
        # Initialize
        self._link = link+"_top/"
        self._topol = topol

        # Create folder
        utils.mkdirp(self._link)


    ###################
    # Private Methods #
    ###################
    def _read(self, link):
        """This function reads a topology file and saves all information and
        data in a dictionary, accessible with the section’s header.

        Parameters
        ----------
        link : string
            Topology-file link

        Returns
        -------
        data : dictionary
            Dictionary containing all data of the topology file
        """
        # Initialize
        data = {}

        # Read data
        is_new = False
        block = ""
        with open(link, "r") as file_in:
            for line in file_in:
                dat = line.split()

                if is_new:
                    if len(dat) > 0:
                        data[block].append(line)
                    else:
                        is_new = False

                elif len(dat) > 0 and dat[0] == "[":
                    block = dat[1]

                    if not dat[1] in data.keys():
                        is_new = True
                        data[block] = []

        return data

    def _itp(self, link, charge):
        """Create an itp file from a GROMACS topology contain everything except

        * defaults
        * atomtypes
        * system
        * molecules

        Parameters
        ----------
        link : string
            Topology file link
        charge : None, float
            Silicon charge for pore grid molecules
        """
        # Initialize
        data = self._read(link)
        dont = ["defaults", "atomtypes", "system", "molecules"]

        itp_link = self._link+link.split("/")[-1].split(".")[0]+".itp"

        # Create itp
        with open(itp_link, "w") as file_out:
            for param in data:
                if not param in dont:
                    file_out.write("[ "+param+" ]\n")
                    for line in data[param]:
                        file_out.write(line)
                    file_out.write("\n")

        # Replace charges in itp files
        if charge is not None:
            utils.replace(itp_link, "CHARGESI", "%8.6f" % charge)


    ##################
    # Public Methods #
    ##################
    def generate_files(self):
        """Generate topology files.
        """
        # Get silicon charge if existent
        charge = self._topol["charge"] if "charge" in self._topol else None

        # Run through topologies
        for top_type in self._topol:
            if not top_type=="charge":
                for file_link in self._topol[top_type]:
                    if top_type=="master":
                        utils.copy(file_link, self._link+"topol.top")
                        utils.copy(file_link, self._link+"topolBackup.top")
                    elif top_type=="top":
                        utils.copy(file_link, self._link+file_link.split("/")[-1])
                    elif top_type=="itp":
                        self._itp(file_link, charge)



################################################################################
# Analyse Class                                                                #
#                                                                              #
"""All necessary function for creating analysis shells."""
################################################################################


import poresim.utils as utils


class Analyze:
    """This class creates shell-files for generating shell file for analysing
    a simulation.

    Parameters
    ----------
    sim_link : string
        Simulation master folder link
    box_link : string
        Simulation box folder link
    """
    def __init__(self, sim_link, box_link, cluster):
        # Initialize
        self._sim_link = sim_link
        self._box_path = box_link
        self._cluster = cluster
        self._box_link = "./" if sim_link == box_link else "./" + \
            box_link.split("/")[-2]+"/"


    ##################
    # Public Methods #
    ##################
    def extract_mol(self, step):
        """This function extracts molecules of a type from a trajectory for
        further analysis using MolDyn.

        Parameters
        ----------
        step : string
            Step to analyze (**nvt**, **npt**, **run**)
        """
        # Set folder names
        folder_step = "../"+step+"/"

        # Set file names
        file_trr = step+".trr"
        file_tpr = step+".tpr"

        # Open file
        utils.mkdirp(self._box_path+"ana")
        with open(self._box_path+"ana/ana.sh", "w") as file_out:
            # Check if backup folder is given
            file_out.write("# Set Todos\n")
            file_out.write("echo \"Load gromacs ...\"; exit;\n")
            file_out.write("echo \"Set MOLECULEINDEX ...\"; exit;\n")
            file_out.write("\n")

            # Extract molecule
            file_out.write("# Extract molecules from trajectory\n")
            file_out.write("declare -A mols\n\n")

            file_out.write("mols[]=""\n\n")

            file_out.write("for key in \"${!mols[@]}\"; do\n")

            out_string = "gmx_mpi trjconv "
            out_string += "-f "+folder_step+file_trr+" "
            out_string += "-s "+folder_step+file_tpr+" "
            out_string += "-o traj_${mols[$key]}.xtc "
            out_string += "-pbc mol "
            out_string += "<<EOF\n"
            out_string += "$key\n"
            out_string += "EOF\n\n"
            file_out.write(out_string)
            
            file_out.write("done\n\n")


            file_out.write("echo \"System "+self._box_link+ " - Finished Extraction ...\"\n\n")

            file_out.write(self._cluster["queuing"]["submit"] + " ana.job")

            



################################################################################
# Actuate Class                                                                #
#                                                                              #
"""All necessary function for running the simulation."""
################################################################################


from re import A
import poresim.utils as utils


class Actuate:
    """This class creates needed shell files for running the simulation using
    GROMACS.

    Parameters
    ----------
    sim_link : string
        Simulation master folder link
    box_link : string
        Simulation box folder link
    cluster: dictionary
        Cluster information dictionary
    job : dictionary
        Job dictionary
    label : string
        Box label
    structure : dictionary
        Box structure dictionary
    """
    def __init__(self, sim_link, box_link, cluster, job, label, structure):
        # Initialize
        self._link = box_link
        self._sim_link = sim_link
        self._box_link = "./" if sim_link == box_link else "./"+box_link.split("/")[-2]+"/"
        self._clr_link = cluster["directory"]+sim_link.split("/")[-2]+"/"
        self._clr_link += "" if sim_link == box_link else box_link.split("/")[-2]+"/"
        self._cluster = cluster
        self._job = job
        self._label = label
        self._is_pore = "PORE" in structure
        self._is_plm = "PLUMED" in structure


    ###################
    # Private methods #
    ###################
    def _equilibration(self):
        """Create equilibration simulation shells for
        * **min** for energy minimization
        * **nvt** for temperature equilibration
        * **npt** for pressure equilibration
        """
        # Set simulation folder descriptors
        sim_min = "min"
        sim_nvt = "nvt"
        sim_npt = "npt"
        sim_all = {sim_min: sim_min, sim_nvt: sim_nvt, sim_npt: sim_npt}

        # Set folder names
        folder_gro = "../_gro/"
        folder_top = "../_top/"
        folder_mdp = "../_mdp/"

        # Set file names
        file_box = "box.gro"
        file_top = "topol.top"
        file_ndx = "index.ndx"

        # Get simulation properties
        np = {step: str(self._job[step]["np"]) for step in self._job}
        ntomp = {step: str(self._job[step]["ntomp"]) if "ntomp" in self._job[step] else "0" for step in self._job}
        nodes = {step: str(self._job[step]["nodes"]) for step in self._job}
        gpu = {step: self._job[step]["gpu"] if "gpu" in self._job[step] else False for step in self._job}
        wall = {step: self._job[step]["wall"] for step in self._job}

        # Set step dependent files
        conf = {"min": folder_gro+file_box,
                "nvt": "../"+sim_min+"/"+sim_min+".gro",
                "npt": "../"+sim_nvt+"/"+sim_nvt+".gro"}
        check = {"min": "",
                    "nvt": "",
                    "npt": "-t ../"+sim_nvt+"/"+sim_nvt+".cpt "}
        forward = {"min": sim_nvt, "nvt": sim_npt, "npt": ""}

        if "npt" not in self._job:
            forward["nvt"] = ""

        # Create job shells
        for step in self._job:
            if not step == "run":
                link_shell = self._link+step+"/"+step+".job"

                utils.mkdirp(self._link+step)
                utils.copy(self._job[step]["file"], link_shell)

                # Simulation options
                utils.replace(link_shell, "SIMULATIONNODES", nodes[step])
                utils.replace(link_shell, "SIMULATIONPROCS", np[step])
                utils.replace(link_shell, "SIMULATIONGPU", "gpu:" + str(gpu[step]) if gpu[step]!=False else "")
                utils.replace(link_shell, "SIMULATIONTIME", wall[step])
                utils.replace(link_shell, "SIMULATIONLABEL", self._label+"_"+step)
                utils.replace(link_shell, "COMMANDCHANGEDIR", "cd "+self._clr_link+step)

                # Grompp
                gro_string = ""
                gro_string += "gmx_mpi grompp "
                gro_string += "-f "+folder_mdp+sim_all[step]+".mdp "
                gro_string += "-c "+conf[step]+" "
                gro_string += check[step]
                gro_string += "-p "+folder_top+file_top+" "
                gro_string += "-n "+folder_gro+file_ndx+" " if self._is_pore else ""
                gro_string += "-po "+sim_all[step]+" "
                gro_string += "-o "+sim_all[step]
                gro_string += " -maxwarn 1" if self._is_pore else ""
                gro_string += "\n"

                # Mdrun
                gro_string += self._cluster["queuing"]["mpi"]+" "
                gro_string += str(int(np[step])*int(nodes[step]))+" " if self._cluster["queuing"]["add_np"] else ""
                gro_string += "gmx_mpi mdrun "
                gro_string += "-s "+sim_all[step]+".tpr "
                gro_string += "-ntomp "+ntomp[step]+" " if not ntomp[step]=="0" else ""
                gro_string += "-plumed ../_gro/plumed.dat " if self._is_plm else ""
                gro_string += "-deffnm "+sim_all[step]+" "
                gro_string += "-dlb no -c -e -g -o -cpo"

                # Insert into file
                utils.replace(link_shell, "COMMANDGROMACS", gro_string)

                # Add forward queuing
                with open(link_shell, "a") as file_out:
                    if not forward[step] == "":
                        file_out.write("\n")
                        file_out.write("sleep 20\n")
                        file_out.write("rm *#\n")
                        file_out.write("cd ../"+forward[step]+"/\n")
                        file_out.write(self._cluster["queuing"]["submit"]+" "+forward[step]+".job"+"\n")

        # Add to master shell
        with open(self._sim_link+"equilibrate.sh", "a") as file_out:
            file_out.write("cd "+self._box_link+sim_min+"/\n")
            file_out.write(self._cluster["queuing"]["submit"]+" "+sim_min+".job\n")
            back = "" if self._box_link == "./" else "../"
            file_out.write("cd "+back+"../\n")
            file_out.write("\n")

    def _simulation(self):
        """Create simulation job files. For the first run - for creating the tpr
        file - run0.job is used. For further simulations, job script run.job is
        created.

        Note that the latter automatically extends the current simulation.
        The breakpoint can be set using the *sim_num* variable in the shell
        file.
        """
        # Set simulation folder descriptor
        sim_nvt = "nvt"
        sim_npt = "npt"
        sim_run = "run"

        # Set folder names
        folder_gro = "../_gro/"
        folder_top = "../_top/"
        folder_mdp = "../_mdp/"

        # Set file names
        file_top = "topol.top"
        file_ndx = "index.ndx"

        # Get last step
        last = "../"+sim_npt+"/"+sim_npt if "npt" in self._job else "../"+sim_nvt+"/"+sim_nvt

        # Get simulation properties
        np = str(self._job["run"]["np"])
        nodes = str(self._job["run"]["nodes"])
        ntomp = str(self._job["run"]["ntomp"]) if "ntomp" in self._job["run"] else "0"
        gpu = self._job["run"]["gpu"] if "gpu" in self._job["run"] else False
        wall = self._job["run"]["wall"]

        # Create job shells
        for i in range(2):
            # Initialize
            is_first = i == 0
            shell_num = "0" if is_first else ""
            sim_num = i+1

            # Create shell
            link_shell = self._link+sim_run+"/"+sim_run+shell_num+".job"

            utils.mkdirp(self._link+sim_run)
            utils.copy(self._job["run"]["file"], link_shell)

            # Change variables
            utils.replace(link_shell, "SIMULATIONNODES", nodes)
            utils.replace(link_shell, "SIMULATIONPROCS", np)
            utils.replace(link_shell, "SIMULATIONGPU", ":gpus=1:exclusive_process" if gpu else "")
            utils.replace(link_shell, "SIMULATIONTIME", wall)
            utils.replace(link_shell, "SIMULATIONLABEL",self._label+"_"+str(sim_num))
            utils.replace(link_shell, "COMMANDCHANGEDIR", "cd "+self._clr_link+sim_run)

            # Grompp
            gro_string = ""
            if is_first:
                gro_string += "gmx_mpi grompp "
                gro_string += "-f "+folder_mdp+sim_run+".mdp "
                gro_string += "-c "+last+".gro "
                gro_string += "-t "+last+".cpt "
                gro_string += "-p "+folder_top+file_top+" "
                gro_string += "-n "+folder_gro+file_ndx+" " if self._is_pore else ""
                gro_string += "-po "+sim_run+" "
                gro_string += "-o "+sim_run+" "
                gro_string += " -maxwarn 1" if self._is_pore else ""
                gro_string += "\n"

            # Mdrun
            gro_string += self._cluster["queuing"]["mpi"]+" "
            gro_string += str(int(np)*int(nodes))+" " if self._cluster["queuing"]["add_np"] else ""
            gro_string += "gmx_mpi mdrun "
            gro_string += "-s "+sim_run+".tpr "
            gro_string += "-ntomp "+ntomp+" " if not ntomp=="0" else ""
            gro_string += "-plumed ../_gro/plumed.dat " if self._is_plm else ""
            gro_string += "-cpi "+sim_run+".cpt " if not is_first else ""
            gro_string += "-deffnm "+sim_run+" "
            gro_string += "-maxh "+str(self._job["run"]["maxh"])+" " if "maxh" in self._job["run"] else ""
            gro_string += "-dlb no -c -e -g -o -cpo"
            gro_string += " -append" if not is_first else ""

            # Insert into file
            utils.replace(link_shell, "COMMANDGROMACS", gro_string)

            # Add forward queuing
            with open(link_shell, "a") as file_out:
                file_out.write("\n")
                file_out.write("sim_num="+str(sim_num)+"\n")

                if not is_first:
                    file_out.write("\n")
                    file_out.write("cp "+sim_run+".job temp.job\n")
                    file_out.write("sed -i \"s/sim_num=$sim_num/sim_num=$((sim_num+1))/\" temp.job\n")
                    file_out.write("sed -i \"s/"+self._label+"_$sim_num/" +
                                  self._label+"_$((sim_num+1))/\" temp.job\n")
                    file_out.write("mv temp.job "+sim_run+".job\n")

                file_out.write("\n")
                file_out.write("sleep 20\n")
                out_string = "if "
                out_string += "[ -f "+sim_run+".cpt ] && "
                out_string += "[ $sim_num -lt "+str(self._job["run"]["runs"])+" ]; "
                out_string += "then "+self._cluster["queuing"]["submit"]+" "+sim_run+".job;"
                out_string += "fi\n"
                file_out.write(out_string)

        # Add to master shell
        with open(self._sim_link+"simulate.sh", "a") as file_out:
            file_out.write("cd "+self._box_link+sim_run+"/\n")
            file_out.write(self._cluster["queuing"]["submit"]+" "+sim_run+"0.job\n")
            back = "" if self._box_link == "./" else "../"
            file_out.write("cd "+back+"../\n")
            file_out.write("\n")

    def _analyze(self):
        """Create analyse job files for poreana.

        """
        # Get simulation properties
        np = str(int(20))
        nodes = str(int(1))
        wall = "10:00:00"

        # Create shell
        utils.mkdirp(self._link+"ana")
        link_shell = self._link+"ana/ana.job"
        utils.copy(self._job["run"]["file"], link_shell)

        # Change variables
        utils.replace(link_shell, "SIMULATIONNODES", nodes)
        utils.replace(link_shell, "SIMULATIONPROCS", np)
        utils.replace(link_shell, "SIMULATIONTIME", wall)
        utils.replace(link_shell, "SIMULATIONLABEL",self._label+"_"+str("ana"))
        utils.replace(link_shell, "COMMANDCHANGEDIR", "cd "+self._clr_link+str("ana"))

        # Insert into file
        utils.replace(link_shell, "COMMANDGROMACS", "python ana.py")

    ##################
    # Public Methods #
    ##################
    def generate_files(self):
        """Generate structure files and shells.
        """
        # Create equilibration files and folders
        self._equilibration()

        # Create simulation files for the production cycle
        if "run" in self._job:
            self._simulation()
        
            # Set ana.job file
            self._analyze()




import sys


if __name__ == "__main__":
    ### Anion ###
    # Set name of ions for ionic liquid simulations
    # an_ion = ""   # Set name "Im"
    # kat_ion = ""  # Set name "FAP"

    # Get user input
    link = sys.argv[1]
    link = link+"/" if not link[-1] == "/" else link

    # Read filled structure
    with open(link+"box.gro", "r") as file_in:
        counter = 0
        data = []
        for line in file_in:
            if counter < 2:
                if counter == 0:
                    title = line
                counter += 1
            elif not len(line.split()) == 3:
                atom = []
                atom.append(int(line[0:5]))
                atom.append(line[5:10])
                atom.append(line[10:15])
                atom.append(int(line[15:20]))
                atom.append(float(line[20:28]))
                atom.append(float(line[28:36]))
                atom.append(float(line[36:44]))

                data.append(atom)
            else:
                box = line

    # Create molecule lists
    mols = {}
    mol_names = []
    last_atom = None
    for atom in data:
        # Create new dictionary entry
        if not atom[1] in mols:
            mols[atom[1]] = []
            mol_names.append(atom[1])

        # Create new molecule
        if last_atom is None or not atom[0]==last_atom:
            mols[atom[1]].append([])

        # Add atom
        mols[atom[1]][-1].append(atom)
        last_atom = atom[0]

    ### Anion ###
    # Same amount of Kat-Ion and An-Ion for ionic liquid simulations
    # if an_ion is not None and kat_ion is not None:
    #     for mol_name in mol_names:
    #         if an_ion in mol_name:
    #             an_ion = mol_name
    #         if kat_ion in mol_name:
    #             kat_ion = mol_name
    #
    #     max_num_ions = min(len(mols[an_ion]), len(mols[kat_ion]))
    #
    #     mols[an_ion] = mols[an_ion][:max_num_ions]
    #     mols[kat_ion] = mols[kat_ion][:max_num_ions]

    # Create new gro file
    with open(link+"box.gro", "w") as file_out:
        # Set title
        file_out.write(title)

        # Number of atoms
        num_atoms = sum([sum([len(mol) for mol in mols[mol_name]]) for mol_name in mol_names])
        file_out.write("%i" % num_atoms+"\n")

        # Atoms
        num_a = 0
        num_m = 0
        for mol_name in mol_names:
            for mol in mols[mol_name]:
                num_m = num_m+1 if num_m < 99999 else 0
                for atom in mol:
                    # Set ids
                    num_a = num_a+1 if num_a < 99999 else 0

                    # Write atom string
                    out_string  = "%5i" % num_m  #  1- 5 (5)    Residue number
                    out_string += atom[1]        #  6-10 (5)    Residue short name
                    out_string += atom[2]        # 11-15 (5)    Atom name
                    out_string += "%5i" % num_a  # 16-20 (5)    Atom number
                    for j in range(3):           # 21-44 (3*8)  Coordinates
                        out_string += "%8.3f" % atom[4+j]

                    file_out.write(out_string+"\n")

        # Box
        file_out.write(box)



################################################################################
# Benchmark Class                                                             #
#                                                                              #
"""All necessary function for running the simulation."""
################################################################################


import poresim.utils as utils

from poresim.simulate import Simulate
from poresim.parameter import Parameter
from poresim.topology import Topology
from poresim.construct import Construct
from poresim.actuate import Actuate


class Benchmark(Simulate):
    """This class extends the simulate class
    :class:`poresim.simulate.Simulate` for benchmarking. It creates a
    benchmark in order to figure out the best possible number of node and
    processor combination to simulate with.

    Depending on the iterator, different aspects will be benchmarked. Currently
    following inputs are possible

    * **nodes** - varying number of nodes at constant CPU number
    * **np** - varying number of CPUs at constant node number

    Parameters
    ----------
    box : Box
        Simulation box object
    np : integer, list
        Number of processors to try
    nodes : integer, list
        Number of nodes to try
    link : string, optional
        Simulation folder link
    iterator : string, optional
        Aspect to be benchmarked
    """
    def __init__(self, box, np, nodes, link="./benchmark", iterator="nodes"):
        # Call super class
        super(Benchmark, self).__init__(link, box)

        # Initialize
        self._box = box
        self._np = np if isinstance(np, list) else [np]
        self._nodes = nodes if isinstance(nodes, list) else [nodes]
        self._iterator = iterator


    ##################
    # Public Methods #
    ##################
    def generate(self):
        """Generate simulation folder.
        """
        # Set parameter dictionary
        params = self._sim_dict["param"]
        params["nvt"]["param"]["NUMBEROFSTEPS"] = 5000
        if "npt" in params:
            params["npt"]["param"]["NUMBEROFSTEPS"] = 5000
        if "run" in params:
            del params["run"]

        # Remove run from job dictionary
        if "run" in self._sim_dict["job"]:
            del self._sim_dict["job"]["run"]

        # Set box specifications
        self._box.set_name("X")
        self._box.set_label("X")
        self._box.set_param(params)

        # Create parameter files
        box_link = self._link+self._box.get_name()+"/"

        # Create folder if multiple boxes exist
        utils.mkdirp(box_link)

        # Create parameter files
        param = Parameter(box_link, self._box.get_param())
        param.generate_files()

        # Create topology files
        topol = Topology(box_link, self._box.get_topol())
        topol.generate_files()

        # Create structure files and shells
        construct = Construct(self._link, box_link, self._box.get_mols(), self._box.get_struct())
        construct.generate_files()

        # Create simulation shells
        actuate = Actuate(self._link, box_link, self._sim_dict["cluster"], self._sim_dict["job"], self._box.get_label(), self._box.get_struct())
        actuate.generate_files()

        # Process iterator input
        if self._iterator == "nodes":
            iterator = self._nodes
        elif self._iterator == "np":
            iterator = self._np

        # Create copy shell after minimization
        with open(self._link+"benchmark.sh", "a") as file_out:
            # Create subfolders for different simulations
            job = self._sim_dict["job"]
            for it in iterator:
                # Set nodes and np
                n = it if iterator == self._np else self._np[0]
                node = it if iterator == self._nodes else self._nodes[0]

                # Set job dictionary
                job["min"]["np"] = n
                job["min"]["nodes"] = node
                job["min"]["wall"] = "01:00:00"

                job["nvt"]["np"] = n
                job["nvt"]["nodes"] = node
                job["nvt"]["wall"] = "01:00:00"

                if "npt" in params:
                    job["npt"]["np"] = n
                    job["npt"]["nodes"] = node
                    job["npt"]["wall"] = "01:00:00"

                # Set subfolder
                self._sim_dict["job"] = job
                self._box.set_name(str(it).zfill(2))
                self._box.set_label("b_"+str(it).zfill(2))
                box_link = self._link+self._box.get_name()+"/"

                # Create parameter files
                param = Parameter(box_link, self._box.get_param())
                param.generate_files()

                # Create simulation shells
                actuate = Actuate(self._link, box_link, self._sim_dict["cluster"], self._sim_dict["job"], self._box.get_label(), self._box.get_struct())
                actuate.generate_files()

                # Write copy to shell
                file_out.write("cp -r X/_gro "+self._box.get_name()+"/_gro\n")
                file_out.write("cp -r X/_top "+self._box.get_name()+"/_top\n")
                file_out.write("cp X/min/* "+self._box.get_name()+"/min/\n\n")

        # Set equilibration master shell to nvt
        utils.replace(self._link+"equilibrate.sh", "min", "nvt")



import os
import sys

import shutil
import unittest

import poresim as ps


class UserModelCase(unittest.TestCase):
    #################
    # Remove Output #
    #################
    @classmethod
    def setUpClass(self):
        if os.path.isdir("tests"):
            os.chdir("tests")

        folder = 'output'
        ps.utils.mkdirp(folder)
        ps.utils.mkdirp(folder+"/temp")
        open(folder+"/temp.txt", 'a').close()

        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)

    #########
    # Utils #
    #########
    def test_utils(self):
        file_link = "output/test/test.txt"

        ps.utils.mkdirp("output/test")

        with open(file_link, "w") as file_out:
            file_out.write("TEST")
        ps.utils.copy(file_link, file_link+"t")
        ps.utils.replace(file_link+"t", "TEST", "DOTA")
        with open(file_link+"t", "r") as file_in:
            for line in file_in:
                self.assertEqual(line, "DOTA\n")

        self.assertEqual(round(ps.utils.mumol_m2_to_mols(3, 100), 4), 180.66)
        self.assertEqual(round(ps.utils.mols_to_mumol_m2(180, 100), 4), 2.989)
        self.assertEqual(round(ps.utils.mmol_g_to_mumol_m2(0.072, 512), 2), 0.14)
        self.assertEqual(round(ps.utils.mmol_l_to_mols(30, 1000), 4), 18.066)
        self.assertEqual(round(ps.utils.mols_to_mmol_l(18, 1000), 4), 29.8904)


    ############
    # Simulate #
    ############
    def test_box(self):
        job = {"min": {"file": "data/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
               "nvt": {"file": "data/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
               "run": {"file": "data/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}

        param = {"min": {"file": "data/pore_min.mdp"},
                 "nvt": {"file": "data/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
                 "run": {"file": "data/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}

        # Boxes
        box = ps.Box("box")
        box.add_box("data/pore.gro")
        box.add_pore("data/pore.yml")
        box.add_mol("EDC", "data/educt.gro", 10)
        box.add_mol("PRD", "data/productmc.gro", 12)
        box.add_mol("BEN", "data/benzene.gro", "fill", auto_dens=500)
        box.add_topol("data/pore.top", "master")
        box.add_topol("data/grid.itp", "top")
        box.add_topol(["data/educt.top", "data/productmc.top", "data/benzene.top"])
        box.add_topol(["data/tms.top", "data/tmsg.itp"])
        box.add_struct("GRO", "data/benzene.gro")
        box.set_job(job)
        box.set_param(param)
        box.add_charge_si(1.28)

        print()
        self.assertEqual(box.add_mol("EDC", "data/educt.gro", 0.1), None)
        self.assertEqual(box.add_mol("EDC", "data/educt.gro", 10, num_atoms="DOTA"), None)
        self.assertEqual(box.add_mol("EDC", "data/educt.gro", 10, auto_dens="DOTA"), None)


    def test_fill_box(self):
        job = {"min": {"file": "data/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
               "nvt": {"file": "data/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
               "run": {"file": "data/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}

        param = {"min": {"file": "data/pore_min.mdp"},
                 "nvt": {"file": "data/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
                 "run": {"file": "data/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}

        # Boxes
        box = ps.Box("box")
        box.add_box("data/box.gro")
        box.add_mol("BEN", "data/benzene.gro", "fill", auto_dens=500)
        box.add_topol("data/benzene.top", "master")
        box.set_job(job)
        box.set_param(param)

        print()


    ############
    # Simulate #
    ############
    def test_sim(self):
        job = {"min": {"file": "data/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
               "nvt": {"file": "data/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
               "run": {"file": "data/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}

        param = {"min": {"file": "data/pore_min.mdp"},
                 "nvt": {"file": "data/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
                 "run": {"file": "data/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}

        cluster = {"address": "user_name@cluster",
                   "directory": "/home/pores/simulation/",
                   "queuing": {"add_np": False, "mpi": "$DO_PARALLEL", "shell": "forhlr.sh", "submit": "sbatch --partition multinode"}}

        box1 = ps.Box("box1")
        box1.add_box("data/pore.gro")
        box1.add_pore("data/pore.yml")
        box1.add_mol("EDC", "data/educt.gro", 10)
        box1.add_mol("PRD", "data/productmc.gro", 12)
        box1.add_mol("BEN", "data/benzene.gro", "fill", auto_dens=500, mass=78.11)
        box1.add_topol("data/pore.top", "master")
        box1.add_topol("data/grid.itp", "top")
        box1.add_topol(["data/educt.top", "data/productmc.top", "data/benzene.top"])
        box1.add_topol(["data/tms.top", "data/tmsg.itp"])
        box1.add_struct("GENERATE", "data/benzene.gro")
        box1.add_struct("PLUMED", "data/benzene.gro")
        box1.set_job(job)
        box1.set_param(param)
        box1.add_charge_si(1.314730)

        box2 = ps.Box("box2", "bxx")
        box2.add_box("data/pore.gro")
        box2.add_pore("data/pore.yml")
        box2.add_mol("EDC", "data/educt.gro", 15)
        box2.add_mol("PRD", "data/productmc.gro", 12)
        box2.add_mol("BEN", "data/benzene.gro", "fill", auto_dens=500, mass=78.11)
        box2.add_topol("data/pore.top", "master")
        box2.add_topol("data/grid.itp", "top")
        box2.add_topol(["data/educt.top", "data/productmc.top", "data/benzene.top"])
        box2.add_topol(["data/tms.top", "data/tmsg.itp"])
        box2.set_job(job)
        box2.set_param(param)
        box2.add_charge_si(1.314730)

        #sim1 = ps.Simulate("output/series", [box1, box2])  # Series
        sim2 = ps.Simulate("output/single", box1)  # Single
        sim3 = ps.Simulate("output/single", box1)

        sim3.add_box(box2)
        sim3.set_sim_dict(sim2.get_sim_dict())
        sim3.set_box(list(sim2.get_box().values()))
        sim3.set_cluster(cluster)

        print()
        #sim1.generate()
        print()
        sim2.generate()


    #############
    # Benchmark #
    #############
    def test_bench(self):
        job = {"min": {"file": "data/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
               "nvt": {"file": "data/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
               "npt": {"file": "data/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
               "run": {"file": "data/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}

        param = {"min": {"file": "data/pore_min.mdp"},
                 "nvt": {"file": "data/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
                 "npt": {"file": "data/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
                 "run": {"file": "data/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}

        box = ps.Box("box")
        box.add_box("data/pore.gro")
        box.add_pore("data/pore.yml")
        box.add_mol("EDC", "data/educt.gro", 10)
        box.add_mol("PRD", "data/productmc.gro", 12)
        box.add_mol("BEN", "data/benzene.gro", "fill", auto_dens=500, mass=78.11)
        box.add_topol("data/pore.top", "master")
        box.add_topol("data/grid.itp", "top")
        box.add_topol(["data/educt.top", "data/productmc.top", "data/benzene.top"])
        box.add_topol(["data/tms.top", "data/tmsg.itp"])
        box.add_struct("GRO", "data/benzene.gro")
        box.set_job(job)
        box.set_param(param)
        box.add_charge_si(1.28)

        bench1 = ps.Benchmark(box, 20, list(range(21)), "output/bench1")
        bench1.set_job(job)
        bench1.set_param(param)
        bench1.generate()

        bench2 = ps.Benchmark(box, 20, list(range(21)), "output/bench2", iterator="np")
        bench2.set_job(job)
        bench2.set_param(param)
        bench2.generate()


if __name__ == '__main__':
    unittest.main(verbosity=2)



################################################################################
# Utils                                                                        #
#                                                                              #
"""Here popular basic methods are noted."""
################################################################################


import os
import fileinput
import yaml
from shutil import copyfile


def load(link):
    with open(link, "r") as file_in:
        return yaml.load(file_in, Loader=yaml.UnsafeLoader)
    
def mkdirp(directory):
    """Create directory if it does not exist.

    Parameters
    ----------
    directory : string
        Directory name
    """
    if not os.path.exists(directory):
        os.makedirs(directory)


def copy(source, target):
    """Copy a specified file to a specified location.

    Parameters
    ----------
    source : string
        Link to requested file
    target : string
        Link to requested new file
    """
    copyfile(source, target)


def replace(file_link, old, new):
    """Replace a given string in a file.

    Parameters
    ----------
    file_link : string
        Link to requested file
    old : string
        String to be replaced in file
    new : string
        New string to be written
    """
    for line in fileinput.input(file_link, inplace=True):
        print(line.rstrip().replace(old, new))


def mumol_m2_to_mols(c, A):
    """Convert the concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    to number of molecules.

    The concentration is given as

    .. math::

        c=\\frac{N}{N_A\\cdot A}

    with surface :math:`A` and avogadro constant
    :math:`N_A=6.022\\cdot10^{23}\\,\\text{mol}^{-1}`. Assuming that the unit of
    the concentration is :math:`\\mu\\text{mol}\\cdot\\text{m}^{-2}` and of the
    surface is :math:`\\text{nm}^2`, the number of molecules is given by

    .. math::

        N=c\\cdot N_A\\cdot A
        =[c]\\cdot10^{-24}\\,\\frac{\\text{mol}}{\\text{nm}^2}\\cdot6.022\\cdot10^{23}\\,\\frac{1}{\\text{mol}}\\cdot[A]\\,\\text{nm}^2

    and thus

    .. math::

        \\boxed{N=0.6022\\cdot[c]\\cdot[A]}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    A : float
        Surface in :math:`\\text{nm}^2`

    Returns
    -------
    N : float
        Number of molecules
    """
    return 0.6022*c*A


def mols_to_mumol_m2(N, A):
    """Convert the number of molecules to concentration in
    :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`.

    The concentration is given as

    .. math::

        c=\\frac{N}{N_A\\cdot A}

    with surface :math:`A` and avogadro constant
    :math:`N_A=6.022\\cdot10^{23}\\,\\text{mol}^{-1}`. Assuming that the unit of
    the concentration is :math:`\\mu\\text{mol}\\cdot\\text{m}^{-2}` and of the
    surface is :math:`\\text{nm}^2`, the number of molecules is given by

    .. math::

        c=\\frac{[N]}{6.022\\cdot10^{23}\\,\\text{mol}^{-1}\\cdot[A]\\,\\text{nm}^2}

    and thus

    .. math::

        \\boxed{c=\\frac{N}{0.6022\\cdot[A]}\\cdot\\frac{\\mu\\text{mol}}{\\text{m}^2}}\\ .

    Parameters
    ----------
    N : int
        Number of molecules
    A : float
        Surface in :math:`\\text{nm}^2`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    """
    return N/0.6022/A


def mmol_g_to_mumol_m2(c, SBET):
    """Convert the concentration :math:`\\frac{\\text{mmol}}{\\text{g}}`
    to :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`.

    This is done by dividing the concentration per gram by the material surface
    per gram property :math:`S_\\text{BET}`

    .. math::

        \\boxed{c_A=\\frac{c_g}{S_\\text{BET}}\\cdot10^3}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{g}}`
    SBET : float
        Material surface in :math:`\\frac{\\text{m}^2}{\\text{g}}`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\mu\\text{mol}}{\\text{m}^2}`
    """
    return c/SBET*1e3


def mmol_l_to_mols(c, V):
    """Convert the concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    to number of molecules.

    The concentration in regard to volume is calculated by

    .. math::

        c=\\frac{N}{N_A\\cdot V}

    with volume :math:`V`. Assuming that the unit of the concentration is
    :math:`\\text{mmol}\\cdot\\text{l}` and of the volume is
    :math:`\\text{nm}^3`, the number of molecules is given by

    .. math::

        N=c\\cdot N_A\\cdot V
        =[c]\\cdot10^{-27}\\,\\frac{\\text{mol}}{\\text{nm}^3}\\cdot6.022\\cdot10^{23}\\,\\frac{1}{\\text{mol}}\\cdot[V]\\,\\text{nm}^3

    and thus

    .. math::

        \\boxed{N=6.022\\cdot10^{-4}\\cdot[c]\\cdot[V]}\\ .

    Parameters
    ----------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    V : float
        Surface in :math:`\\text{nm}^3`

    Returns
    -------
    N : float
        Number of molecules
    """
    return 6.022e-4*c*V


def mols_to_mmol_l(N, V):
    """Convert the number of molecules to concentration in
    :math:`\\frac{\\text{mmol}}{\\text{l}}`.

    The concentration in regard to volume is calculated by

    .. math::

        c=\\frac{N}{N_A\\cdot V}

    with volume :math:`V`. Assuming that the unit of the concentration is
    :math:`\\text{mmol}\\cdot\\text{l}` and of the volume is
    :math:`\\text{nm}^3`, the number of molecules is given by

    .. math::

        c=\\frac{N}{6.022\\cdot10^{23}\\cdot[V]\\,\\text{nm}^3}

    and thus

    .. math::

        \\boxed{c=\\frac{N}{6.022\\times10^{-4}\\cdot[V]}\\cdot\\frac{\\text{mmol}}{\\text{l}}}\\ .

    Parameters
    ----------
    N : int
        Number of molecules
    V : float
        Surface in :math:`\\text{nm}^3`

    Returns
    -------
    c : float
        Concentration in :math:`\\frac{\\text{mmol}}{\\text{l}}`
    """
    return N/6.022e-4/V



################################################################################
# Construct Class                                                              #
#                                                                              #
"""All necessary function for creating the finished simulation box."""
################################################################################


import os

import poresim.utils as utils
import numpy as np

class Construct:
    """This class creates shell-files for generating the and filling the
    simulation box using GROMACS.

    Parameters
    ----------
    sim_link : string
        Simulation master folder link
    box_link : string
        Simulation box folder link
    struct : dictionary
        Structure dictionary
    """
    def __init__(self, sim_link, box_link, mols, struct):
        # Initialize
        self._sim_link = sim_link
        self._box_path = box_link
        self._box_link = "./" if sim_link == box_link else "./"+box_link.split("/")[-2]+"/"
        self._mols = mols
        self._struct = struct
        if "PORE" in struct:
            self._pore_props = utils.load(struct["PORE"])

    ###################
    # Private Methods #
    ###################
    def _topol_index(self, file_out, path):
        """Helper function for updating the topology and creating an index file,
        if the simulated system is a pore.

        Parameters
        ----------
        file_out : File
            File object to write in
        path : String
            Simulation root path
        """
        # Set folder names
        folder_gro = path+"_gro/"
        folder_top = path+"_top/"

        # Set file names
        file_box = "box.gro"
        file_top = "topol.top"
        file_ndx = "index.ndx"

        # Add number of residues to topology
        file_out.write("# Update Topology\n")
        for mol in self._mols:
            file_out.write("count"+mol+"=$(($(grep -c \""+mol+"\" "+folder_gro+file_box+")/"+str(self._mols[mol][1])+"))\n")
            file_out.write("echo \""+mol+" \"$count"+mol+" >> "+folder_top+file_top+"\n")

        file_out.write("echo \"System "+self._box_link+" - Updated topology ...\"\n\n")

        # Create index file
        if "PORE" in self._struct:
            file_out.write("# Create Index\n")
            out_string = "gmx_mpi make_ndx "
            out_string += "-f "+folder_gro+file_box+" "
            out_string += "-o "+folder_gro+file_ndx
            out_string += " >> logging.log 2>&1 <<EOF\n"
            out_string += "0 & a SI1 OM1\n"
            out_string += "q\n"
            out_string += "EOF\n"
            file_out.write(out_string)

            file_out.write("echo \"System "+self._box_link+" - Created pore index file ...\"\n")

    def _pos_dat(self):
        """
        Helper function to create position files for Gromacs to insert molecules in the reservoir, in the pore or in a specify area of a box system
        """
        for mol in self._mols:
            # Position file for reservoir
            if (self._mols[mol][0]=="fill" and not self._mols[mol][6] and self._mols[mol][4] in ["res", "both"]) and "PORE" in self._struct:
                num = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._pore_props["system"]["dimensions"][0]*self._pore_props["system"]["dimensions"][1]*self._pore_props["system"]["reservoir"]*2*0.5)
                with open(self._box_path +"_gro/" + "position_{}.dat".format(mol), "w") as file_out:
                    for i in range(int(num/2)):
                        out_string = str(self._pore_props["system"]["dimensions"][0]/2) + " "
                        out_string += str(self._pore_props["system"]["dimensions"][1]/2) + " "
                        out_string += str(self._pore_props["system"]["reservoir"]/2) + "\n"
                        file_out.write(out_string)
                    for i in range(int(num/2)):
                        out_string = str(self._pore_props["system"]["dimensions"][0]/2) + " "
                        out_string += str(self._pore_props["system"]["dimensions"][1]/2) + " "
                        out_string += str(self._pore_props["system"]["dimensions"][2] - self._pore_props["system"]["reservoir"]/2) + "\n"
                        file_out.write(out_string)
                    file_out.close()
                # Position files for pore
                for pore_id in self._pore_props.keys():
                    if pore_id[:5]=="shape":
                        if (self._pore_props[pore_id]["parameter"]["central"]==[0,0,1]) and (self._mols[mol][4] in ["pore", "both"]): 
                            num_pore = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*np.pi*self._pore_props[pore_id]["diameter"]**2/4*(self._pore_props[pore_id]["parameter"]["length"]))
                            with open(self._box_path +"_gro/" + "position_{}_{}.dat".format(pore_id,mol), "w") as file_out:
                                for i in range(num_pore):
                                    out_string = str(self._pore_props[pore_id]["parameter"]["centroid"][0]) + " "
                                    out_string += str(self._pore_props[pore_id]["parameter"]["centroid"][1]) + " "
                                    out_string += str(self._pore_props["system"]["reservoir"] + self._pore_props[pore_id]["parameter"]["centroid"][2]) + "\n"
                                    file_out.write(out_string)
                                file_out.close()
            elif (self._mols[mol][0]=="fill" and not self._mols[mol][6] and self._mols[mol][4] in ["pore"]) and "PORE" in self._struct:
                for pore_id in self._pore_props.keys():
                        if pore_id[:5]=="shape":
                            if (self._pore_props[pore_id]["parameter"]["central"]==[0,0,1]) and (self._mols[mol][4] in ["pore", "both"]): 
                                num_pore = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*np.pi*self._pore_props[pore_id]["diameter"]**2/4*(self._pore_props[pore_id]["parameter"]["length"]))
                                with open(self._box_path +"_gro/" + "position_{}_{}.dat".format(pore_id,mol), "w") as file_out:
                                    for i in range(num_pore):
                                        out_string = str(self._pore_props[pore_id]["parameter"]["centroid"][0]) + " "
                                        out_string += str(self._pore_props[pore_id]["parameter"]["centroid"][1]) + " "
                                        out_string += str(self._pore_props["system"]["reservoir"] + self._pore_props[pore_id]["parameter"]["centroid"][2]) + "\n"
                                        file_out.write(out_string)
                                    file_out.close()
            # Position file for put a specific number of molecules in pore or reservoir 
            elif self._mols[mol][0]!="fill" and not self._mols[mol][5] and not self._mols[mol][6] and "PORE" in self._struct:   
                j=0
                # Check how much molecules have to set in pore or in reservoir if a number is given
                for pore_id in self._pore_props.keys():
                        if pore_id[:5]=="shape":
                            j = j + 1 
                if (self._mols[mol][4]=="both"):
                    num = [int(self._mols[mol][0]/(j+2)) for i in range(j+2)]

                    if sum(num) != self._mols[mol][0]:
                        num[-1] = num[-1] + abs(self._mols[mol][0]-sum(num))

                elif (self._mols[mol][4]=="res"):
                    num = [int(self._mols[mol][0]/(2)) for i in range(2)]
                    if sum(num) != self._mols[mol][0]:
                        num[-1] = num[-1] + abs(self._mols[mol][0]-sum(num))

                elif (self._mols[mol][4]=="pore"):
                    num = [int(self._mols[mol][0]/(j)) for i in range(j)]
                    if sum(num) != self._mols[mol][0]:
                        num[-1] = num[-1] + abs(self._mols[mol][0]-sum(num))
                elif (self._mols[mol][4]=="box"):
                    num = [int(self._mols[mol][0]/(j)) for i in range(j)]
                    if sum(num) != self._mols[mol][0]:
                        num[-1] = num[-1] + abs(self._mols[mol][0]-sum(num))

                # Generate the position files for the number of molecules to set in the pores
                if self._mols[mol][4] in ["pore", "both"]:
                    for pore_id,j in zip(self._pore_props.keys(),range(j)):
                        if pore_id[:5]=="shape":
                            if self._pore_props[pore_id]["parameter"]["central"]==[0,0,1]:
                                with open(self._box_path +"_gro/" + "position_{}_{}.dat".format(pore_id,mol), "w") as file_out:
                                    for i in range(num[j]):
                                        out_string = str(self._pore_props[pore_id]["parameter"]["centroid"][0]) + " "
                                        out_string += str(self._pore_props[pore_id]["parameter"]["centroid"][1]) + " "
                                        out_string += str(self._pore_props["system"]["reservoir"] + self._pore_props[pore_id]["parameter"]["centroid"][2]) + "\n"
                                        file_out.write(out_string)
                                    file_out.close()

                # Generate the position files for the number of molecules to set in reservoir
                if self._mols[mol][4] in ["res", "both"]:
                    with open(self._box_path +"_gro/" + "position_{}.dat".format(mol), "w") as file_out:
                        for i in range(num[0]):
                            out_string = str(self._pore_props["system"]["dimensions"][0]/2) + " "
                            out_string += str(self._pore_props["system"]["dimensions"][1]/2) + " "
                            out_string += str(self._pore_props["system"]["reservoir"]/2) + "\n"
                            file_out.write(out_string)
                        for i in range(num[-1]):
                            out_string = str(self._pore_props["system"]["dimensions"][0]/2) + " "
                            out_string += str(self._pore_props["system"]["dimensions"][1]/2) + " "
                            out_string += str(self._pore_props["system"]["dimensions"][2] - self._pore_props["system"]["reservoir"]/2) + "\n"
                            file_out.write(out_string)
                        file_out.close()

                # If slit pore you can set the molecules on wall 
                elif self._mols[mol][4]=="wall":
                    for pore_id,j in zip(self._pore_props.keys(),range(j)):
                        if pore_id[:5]=="shape":
                                with open(self._box_path +"_gro/" + "position_{}_{}.dat".format(pore_id,mol), "w") as file_out:
                                    for i in range(int(self._mols[mol][0]/2)):
                                        out_string = str(self._pore_props[pore_id]["parameter"]["centroid"][0]) + " "
                                        out_string += str(self._pore_props[pore_id]["parameter"]["centroid"][1]-self._pore_props[pore_id]["parameter"]["height"]/2*0.75) + " "
                                        out_string += str(self._pore_props["system"]["reservoir"] + self._pore_props[pore_id]["parameter"]["centroid"][2]) + "\n"
                                        file_out.write(out_string)
                                    for i in range(int(self._mols[mol][0]/2)):
                                        out_string = str(self._pore_props[pore_id]["parameter"]["centroid"][0]) + " "
                                        out_string += str(self._pore_props[pore_id]["parameter"]["centroid"][1]+self._pore_props[pore_id]["parameter"]["height"]/2*0.75) + " "
                                        out_string += str(self._pore_props["system"]["reservoir"] + self._pore_props[pore_id]["parameter"]["centroid"][2]) + "\n"
                                        file_out.write(out_string)

            # Position file for box system 
            elif self._mols[mol][0]=="fill" and self._mols[mol][5] and not self._mols[mol][4]=="wall":
                for area,i in zip(self._mols[mol][5], range(len(self._mols[mol][5]))):
                    num = int((self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._mols[mol][6][0]*self._mols[mol][6][1]*(area[1]-area[0]))*0.8)
                    with open(self._box_path +"_gro/" + "position_{}_area{}.dat".format(mol,i), "w") as file_out:
                        for i in range(num):
                            out_string = str(self._mols[mol][6][0]/2) + " "
                            out_string += str(self._mols[mol][6][1]/2) + " "
                            out_string += str((area[1]+area[0])/2) + "\n"
                            file_out.write(out_string)
                        file_out.close()
            

            elif self._mols[mol][0]=="fill" and not self._mols[mol][5] and not self._mols[mol][4]=="wall":
                num = int((self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._mols[mol][6][0]*self._mols[mol][6][1]*self._mols[mol][6][2])*0.8)
                with open(self._box_path +"_gro/" + "position_{}.dat".format(mol), "w") as file_out:
                    for i in range(num):
                        out_string = str(self._mols[mol][6][0]/2) + " "
                        out_string += str(self._mols[mol][6][1]/2) + " "
                        out_string += str((self._mols[mol][6][2])/2) + "\n"
                        file_out.write(out_string)
                file_out.close()
            
            elif self._mols[mol][0]!="fill" and not self._mols[mol][5]:
                if not self._mols[mol][6]:
                    print("If you fill a system with on molecule specify in add_mol for every molecule the box dimension")
                    exit()
                num = self._mols[mol][0]
                with open(self._box_path +"_gro/" + "position_{}.dat".format(mol), "w") as file_out:
                    for i in range(num):
                        out_string = str(self._mols[mol][6][0]/2) + " "
                        out_string += str(self._mols[mol][6][1]/2) + " "
                        out_string += str((self._mols[mol][6][2])/2) + "\n"
                        file_out.write(out_string)
                file_out.close()

            # Position file for box system to put molecules in a certain area
            elif self._mols[mol][0]!="fill" and self._mols[mol][5]:
                num = self._mols[mol][0]
                for area,i in zip(self._mols[mol][5], range(len(self._mols[mol][5]))):
                    with open(self._box_path +"_gro/" + "position_{}_area{}.dat".format(mol,i), "w") as file_out:
                        for i in range(int(num/len(self._mols[mol][5]))):
                            out_string = str(self._mols[mol][6][0]/2) + " "
                            out_string += str(self._mols[mol][6][1]/2) + " "
                            out_string += str((area[1]+area[0])/2) + "\n"
                            file_out.write(out_string)
                    file_out.close()


    def _structure(self):
        """Create a shell file for constructing and filling the simulation
        box using GROMACS. Additionally, the master topology file is updated
        with the number of added molecules. In case a pore simulation is
        intended the needed index-file containing a new group for SI and OM is
        generated.
        """
        # Set folder names
        folder_gro = self._box_link+"_gro/"
        folder_fill = self._box_link+"_fill/"

        # Set file names
        file_box = "box.gro"

        # Open file
        with open(self._sim_link+"construct.sh", "a") as file_out:
            # Create box label
            file_out.write("#"*(12+len(self._box_link))+"\n")
            file_out.write("# Process "+self._box_link+" #\n")
            file_out.write("#"*(12+len(self._box_link))+"\n")
            file_out.write("echo \"Load gromacs ...\"; exit;\n")
            if "fill" in [self._mols[mol][0] for mol in self._mols]:
                file_out.write("echo \"Set ions names in sort script if necessary ...\"; exit;\n")
            
            # First lines for standard gromacs command
            gmx_standard =  "gmx_mpi insert-molecules " + "-f "+folder_gro+file_box+" " +  "-o "+folder_gro+file_box+" "
            
            # Loop over the molecules which you want to insert into the system
            for mol in self._mols:
                file_out.write("\n####### " + mol + " #########\n")
                # Fill box
                if "PORE" in self._struct:
                    if (self._mols[mol][0]=="fill") and (self._mols[mol][3]!=None):
                        num = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._pore_props["system"]["dimensions"][0]*self._pore_props["system"]["dimensions"][0]*self._pore_props["system"]["reservoir"]*2)
                    else:
                        num = 0
                    if (self._mols[mol][4] in ["res","both"]):
                        # Fill reservoir 
                        file_out.write("# Fill Reservoir " + mol +" \n")
                        out_string = gmx_standard
                        out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                        out_string += "-dr "+ str(self._pore_props["system"]["dimensions"][0]/2) + " " + str(self._pore_props["system"]["dimensions"][1]/2) + " " + str(self._pore_props["system"]["reservoir"]/2) + " "
                        out_string += "-ip "+ folder_gro + "position_{}.dat".format(mol)  +" "
                        out_string += "-nmol "+str(int(self._mols[mol][0])) if not self._mols[mol][0]=="fill" else "-nmol "+str(num) 
                        for key,value in self._mols[mol][-1].items():
                            out_string +=  " "
                            out_string += key + " " + str(value) + " "
                        out_string += " >> logging.log 2>&1\n"
                        file_out.write(out_string)
                        file_out.write("echo \"Filled reservoir with " + mol + " ...\"\n\n")

                    # Fill pore area
                    if self._mols[mol][4]=="pore" or self._mols[mol][4]=="both":
                        file_out.write("# Fill Pore " + mol +" \n")
                        for pore_id in self._pore_props.keys():
                            if pore_id[:5]=="shape":
                                if self._pore_props[pore_id]["parameter"]["central"]==[0,0,1]:
                                    if (self._mols[mol][0]=="fill" and self._mols[mol][3]!=None ):
                                        if self._pore_props[pore_id]["shape"]=="SLIT":
                                            num_pore = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._pore_props["system"]["dimensions"][0]*self._pore_props[pore_id]["diameter"]*(self._pore_props[pore_id]["parameter"]["length"]))
                                        else:
                                            num_pore = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*np.pi*self._pore_props[pore_id]["diameter"]**2/4*(self._pore_props[pore_id]["parameter"]["length"]))
                                    else:
                                        num_pore = 0
                                    if self._mols[mol][4]!="res":
                                        out_string = gmx_standard
                                        out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                                        if self._pore_props[pore_id]["shape"]=="SLIT":
                                            out_string += "-dr "+ str(0.90*self._pore_props["system"]["dimensions"][0]/2) + " " + str(0.90*self._pore_props[pore_id]["diameter"]/2) + " " + str((0.9*self._pore_props[pore_id]["parameter"]["length"])/2) + " "
                                        else:
                                            out_string += "-dr "+ str(0.50*np.sqrt(0.9 *self._pore_props[pore_id]["diameter"]**2)/2) + " " + str(0.50*np.sqrt(0.9 *self._pore_props[pore_id]["diameter"]**2)/2) + " " + str((0.9*self._pore_props[pore_id]["parameter"]["length"])/2) + " "
                                        out_string += "-ip "+ folder_gro + "position_{}_{}.dat".format(pore_id,mol)  +" " 
                                        out_string += "-nmol "+str(int(self._mols[mol][0])) if not self._mols[mol][0]=="fill" else "-nmol "+str(num_pore) + "  "
                                        for key,value in self._mols[mol][-1].items():
                                            out_string +=  " "
                                            out_string += key + "  " + str(value) + " "
                                        out_string += " >> logging.log 2>&1\n"
                                        file_out.write(out_string)
                        file_out.write("echo \"Filled pore with " + mol + " ...\"\n\n")
                    
                    # Fill wall
                    elif self._mols[mol][4] == "wall":
                        file_out.write("# Fill Pore Wall " + mol +" \n")
                        for pore_id in self._pore_props.keys():
                            if pore_id[:5]=="shape":
                                if self._pore_props[pore_id]["parameter"]["central"]==[0,0,1]:
                                    out_string = gmx_standard
                                    out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                                    out_string += "-dr "+ str(0.9*self._pore_props["system"]["dimensions"][0]/2) + " " + str(self._pore_props[pore_id]["diameter"]/2*0.1) + " " + str((self._pore_props[pore_id]["parameter"]["length"])/2) + " "
                                    out_string += "-ip "+ folder_gro + "position_{}_{}.dat".format(pore_id,mol)  +" " 
                                    out_string += "-nmol "+str(int(self._mols[mol][0])) if not self._mols[mol][0]=="fill" else "-nmol "+str(num_pore) + "  "
                                    for key,value in self._mols[mol][-1].items():
                                        out_string +=  " "
                                        out_string += key + "  " + str(value) + " "
                                    out_string += " >> logging.log 2>&1\n"
                                    file_out.write(out_string)
                                file_out.write("echo \"Filled " + pore_id + " " +  mol + " ...\"\n\n")
                # If only box system
                else:
                    # Fill box system in a certain area
                    file_out.write("# Fill Box\n")
                    if self._mols[mol][5]:
                        for area,i in zip(self._mols[mol][5], range(len(self._mols[mol][5]))):
                            if self._mols[mol][0]=="fill":
                                num = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._mols[mol][6][0]*self._mols[mol][6][1]*(area[1]-area[0]))
                            else:
                                num = self._mols[mol][0]
                            out_string = gmx_standard
                            out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                            out_string += "-dr "+ str(self._mols[mol][6][0]/2) + " " + str(self._mols[mol][6][1]/2) + " " + str((area[1]-area[0])/2) + " "
                            out_string += "-ip "+ folder_gro + "position_{}_area{}.dat".format(mol,i)  +" " 
                            out_string += "-nmol "+str(int(self._mols[mol][0])) if not self._mols[mol][0]=="fill" else "-nmol "+str(num) + " "
                            for key,value in self._mols[mol][-1].items():
                                out_string += " "
                                out_string += key + " " + str(value) + " "
                            out_string += " >> logging.log 2>&1\n"
                            file_out.write(out_string)   
                    else: 
                        # Fill box with density or specific number of molecules
                        if self._mols[mol][0]=="fill":
                            num = int(self._mols[mol][2]/self._mols[mol][3]/10*6.022*self._mols[mol][6][0]*self._mols[mol][6][1]*self._mols[mol][6][2])
                        else:
                            num = self._mols[mol][0]
                        out_string = gmx_standard
                        out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                        if self._mols[mol][0]=="fill":
                            out_string += "-dr "+ str(self._mols[mol][6][0]/2) + " " + str(self._mols[mol][6][1]/2) + " " + str((self._mols[mol][6][2])/2) + " "
                            out_string += "-ip "+ folder_gro + "position_{}.dat".format(mol)  +" " 
                        out_string += "-nmol "+str(int(num))
                        for key,value in self._mols[mol][-1].items():
                            out_string += " "
                            out_string += key + " " + str(value) + " "
                        out_string += " >> logging.log 2>&1\n"
                        file_out.write(out_string)

            #if "fill" in [self._mols[mol][0] for mol in self._mols]:
            file_out.write("python "+folder_fill+"sort.py "+folder_gro+"\n")
            file_out.write("echo \"System "+self._box_link+" - Filled simulation box ...\"\n\n")

            # Update topology and create index
            self._topol_index(file_out, self._box_link)

            # Remove log and backup
            file_out.write("rm "+folder_gro+"*#\n")
            file_out.write("rm logging.log\n\n")

    def _fill(self):
        """This function creates a shell file for continuously filling a pore
        simulation manually. The nvt and min equilibration of the last step are
        moved to a backup folder and the last structure is moved to the
        structure folder for the new construction and simulation.
        """
        # Set simulation folder descriptors
        sim_ana = "ana"
        sim_min = "min"
        sim_nvt = "nvt"

        # Set folder names
        folder_fill = "./"
        folder_gro = "../_gro/"
        folder_top = "../_top/"
        folder_ana = "../"+sim_ana+"/"
        folder_min = "../"+sim_min+"/"
        folder_nvt = "../"+sim_nvt+"/"

        # Set file names
        file_box = "box.gro"
        file_top = "topol.top"
        file_t_b = "topolBackup.top"
        file_ndx = "index.ndx"

        # Open file
        with open(self._box_path+"_fill/fill.sh", "w") as file_out:
            # Check if backup folder is given
            file_out.write("# Create Todos\n")
            file_out.write("echo \"Load gromacs ...\"; exit;\n")
            if not all(x is None for x in [self._mols[mol][2] for mol in self._mols]):
                file_out.write("echo \"Load gromacs in Backup...\"; exit;\n")
            file_out.write("echo \"Set ions names in sort script if necessary ...\"; exit;\n")
            file_out.write("\n")

            # Add folder number
            file_out.write("# Set folder number\n")
            file_out.write("fill_num=1\n\n")

            # Backup simulation
            file_out.write("# Backup Simulation\n")
            out_string = "mkdir "+folder_fill+"$fill_num\n"
            out_string += "mv "+folder_gro+file_box+" "+folder_fill+"$fill_num\n"
            out_string += "mv "+folder_top+file_top+" "+folder_fill+"$fill_num\n"
            if "PORE" in self._struct:
                out_string += "mv "+folder_gro+file_ndx+" "+folder_fill+"$fill_num\n"
            out_string += "cp "+folder_nvt+sim_nvt+".gro "+folder_gro+file_box+"\n"
            out_string += "cp "+folder_top+file_t_b+" "+folder_top+file_top+"\n"
            out_string += "mv "+folder_min+" "+folder_fill+"$fill_num\n"
            out_string += "mv "+folder_nvt+" "+folder_fill+"$fill_num\n"
            out_string += "mkdir "+folder_min+"\n"
            out_string += "mkdir "+folder_nvt+"\n"
            out_string += "cp "+folder_fill+"$fill_num/"+sim_min+"/"+sim_min+".job"+" "+folder_min+"\n"
            out_string += "cp "+folder_fill+"$fill_num/"+sim_nvt+"/"+sim_nvt+".job"+" "+folder_nvt+"\n"
            file_out.write(out_string)
            file_out.write("echo \"System "+self._box_link+" - Backed up equilibration ...\"\n\n")

            # Backup Analysis in case of automatic density
            if not all(x is None for x in [self._mols[mol][2] for mol in self._mols]):
                file_out.write("# Backup Analysis\n")
                out_string = "mv "+folder_ana+" "+folder_fill+"$fill_num\n"
                out_string += "mkdir "+folder_ana+"\n"
                out_string += "cp "+folder_fill+"$fill_num/ana/ana.* "+folder_ana+"\n"
                file_out.write(out_string)
                file_out.write("echo \"System "+self._box_link+" - Backed up analysis ...\"\n\n")

            # Fill box
            file_out.write("# Refill Box\n")
            for mol in self._mols:
                if self._mols[mol][0]=="fill" and self._mols[mol][5]:
                    for area,i in zip(self._mols[mol][5], range(len(self._mols[mol][5]))):
                        out_string = "gmx_mpi insert-molecules "
                        out_string += "-f "+folder_gro+file_box+" "
                        out_string += "-o "+folder_gro+file_box+" "
                        out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                        out_string += "-try 1000 "
                        out_string += "-scale 0.47 "
                        out_string += "-dr "+ str(self._mols[mol][6][0]/2) + " " + str(self._mols[mol][6][1]/2) + " " + str((area[1]-area[0])/2) + " "
                        out_string += "-ip "+ folder_gro + "position_{}_area{}.dat".format(mol,i)  +" " 
                        out_string += "-nmol "
                        out_string += str(10000) if self._mols[mol][2] is None else ("FILLDENS_" + mol)
                        out_string += " >> logging.log 2>&1\n" 
                        file_out.write(out_string)
                elif self._mols[mol][0]=="fill" and not self._mols[mol][5]:
                    out_string = "gmx_mpi insert-molecules "
                    out_string += "-f "+folder_gro+file_box+" "
                    out_string += "-o "+folder_gro+file_box+" "
                    out_string += "-ci "+folder_gro+self._struct[mol].split("/")[-1]+" "
                    out_string += "-try 1000 "
                    out_string += "-scale 0.47 "
                    if "PORE" in self._struct:
                        out_string += "-dr "+ str(self._pore_props["system"]["dimensions"][0]/2) + " " + str(self._pore_props["system"]["dimensions"][1]/2) + " " + str(self._pore_props["system"]["reservoir"]/2) + " "
                    else:
                        out_string += "-dr "+ str(self._mols[mol][6][0]/2) + " " + str(self._mols[mol][6][1]/2) + " " + str(self._mols[mol][6][2]/2) + " "
                    out_string += "-ip "+ folder_gro + "position_{}.dat".format(mol)  +" " 
                    out_string += "-nmol "
                    out_string += 0 if self._mols[mol][2] is None else ("FILLDENS_" + mol)
                    out_string += " >> logging.log 2>&1\n"
                    file_out.write(out_string)
            file_out.write("python sort.py "+folder_gro+"\n")
            file_out.write("echo \"System "+self._box_link+" - Refilled simulation box ...\"\n\n")

            # Update topology and create index
            self._topol_index(file_out, "../")
            file_out.write("\n")

            # Remove log and backup
            file_out.write("# Remove logs\n")
            file_out.write("rm "+folder_gro+"*#\n")
            file_out.write("rm logging.log\n\n")

            # Step a folder number forward
            file_out.write("# Step fill folder number\n")
            file_out.write("cp fill.sh temp.sh\n")
            file_out.write("sed -i \"s/fill_num=$fill_num/fill_num=$((fill_num+1))/\" temp.sh\n")
            if not all(x is None for x in [self._mols[mol][2] for mol in self._mols]):
                file_out.write("sed -i \"s/fill_num=$fill_num/fill_num=$((fill_num+1))/\" fillBackup.sh\n")
            file_out.write("mv temp.sh fill.sh\n")


    ##################
    # Public Methods #
    ##################
    def generate_files(self):
        """Generate structure files and shells.
        """
        # Create construction shell file
        self._structure()

        # Create structure folder
        utils.mkdirp(self._box_path+"_gro")

        # Copy structure files
        for mol in self._struct:
            file_link = self._struct[mol]
            if mol=="BOX":
                utils.copy(file_link, self._box_path+"_gro/"+"box.gro")
            elif mol=="GENERATE":
                utils.copy(file_link, self._box_path+"_gro/"+"generate.sh")
            elif mol=="PLUMED":
                utils.copy(file_link, self._box_path+"_gro/"+"plumed.dat")
            else:
                utils.copy(file_link, self._box_path+"_gro/"+file_link.split("/")[-1])
  
        # If molecules have to put in a specific area of a pore system 
        if "PORE" in self._struct:
            if ("wall" or "pore" in [self._mols[mol][4] for mol in self._mols]):
                if not "box" in [self._mols[mol][4] for mol in self._mols]:
                    self._pos_dat()

        # If molecules have to set in a specific area of a box system
        for mol in self._mols:
            if self._mols[mol][5]:
                self._pos_dat()

        # Pore simulation that needs to be filled
        if "fill" in [self._mols[mol][0] for mol in self._mols]:
            # Create filling backup folder
            utils.mkdirp(self._box_path+"_fill")

            # Create position file
            self._pos_dat()

            # Create shell files
            self._fill()

            # Create fill backup for automatic filling
            if not all(x is None for x in [self._mols[mol][2] for mol in self._mols]):
                utils.copy(self._box_path+"_fill/fill.sh", self._box_path+"_fill/fillBackup.sh")

            utils.copy(os.path.split(__file__)[0]+"/templates/sort.py", self._box_path+"_fill/"+"sort.py")



################################################################################
# Simulate Class                                                               #
#                                                                              #
"""Create simulation boxes with all necessary files."""
################################################################################


import os
import yaml
import shutil
from jinja2 import Template

import poresim.utils as utils

from poresim.box import Box
from poresim.parameter import Parameter
from poresim.topology import Topology
from poresim.construct import Construct
from poresim.actuate import Actuate
from poresim.analyze import Analyze


class Simulate:
    """This class generates the simulation folder for a series of simulations.

    The output of this class can then be uploaded to a computing cluster.

    Parameters
    ----------
    link : string, optional
        Simulation folder link
    box : Box, list, optional
        Simulation box object, a series of those
    """
    def __init__(self, link="./simulation/", box=None):
        # Initialize
        self._link = link if link[-1] == "/" else link+"/"
        self._config = self._link+"/"+"config.yml"

        # Create simulation folder if not existent
        utils.mkdirp(self._link)

        # Create simulation config file if not existent
        if not os.path.isfile(self._config):
            # Get package directory
            package_dir = os.path.split(__file__)[0]+"/"

            # Job file
            file_in = package_dir+"templates/config.yml"
            file_out = self._config
            shutil.copy(file_in, file_out)

        # Load simulation config file
        with open(self._config, "r") as stream:
            self._sim_dict = yaml.load(stream, Loader=yaml.FullLoader)

            # Create box objects
            boxes = {}
            for i in self._sim_dict["box"]:
                temp = Box()
                temp.set_sim_dict(self._sim_dict["box"][i])
                boxes[i] = temp
            self._sim_dict["box"] = boxes

        # Add boxes to simulation dictionary
        if box is not None:
            boxes = box if isinstance(box, list) else [box]
            self._sim_dict["box"] = {i:boxes[i] for i in range(len(boxes))}
            self._update_config()


    ###################
    # Private Methods #
    ###################
    def _update_config(self):
        """Update simulation config file.
        """
        sim_dict = self._sim_dict.copy()

        sim_dict["box"] = {box_id:sim_dict["box"][box_id].get_sim_dict() for box_id in sim_dict["box"]}

        with open(self._config, "w") as file_out:
            file_out.write(yaml.dump(sim_dict))


    ##################
    # Public Methods #
    ##################
    def add_box(self, box):
        """Add simulation box to simulation dictionary.

        Parameters
        ----------
        box : Box, list
            Simulation box or a list of those
        """
        boxes = box if isinstance(box, list) else [box]
        for box in boxes:
            self._sim_dict["box"][len(boxes)] = box
        self._update_config()

    def generate(self):
        """Generate simulation folder.
        """
        # Run through boxes
        for box_id in self._sim_dict["box"]:
            box = self._sim_dict["box"][box_id]
            box_link = self._link+box.get_name()+"/" if len(self._sim_dict["box"])>1 else self._link

            # Create folder if multiple boxes exist
            utils.mkdirp(box_link)

            # Create parameter files
            params = self._sim_dict["param"] if box.get_param() is None else box.get_param()
            param = Parameter(box_link, params)
            param.generate_files()

            # Create topology files
            topol = Topology(box_link, box.get_topol())
            topol.generate_files()

            # Create structure files and shells
            construct = Construct(self._link, box_link, box.get_mols(), box.get_struct())
            construct.generate_files()

            # Create simulation shellsboxes[sim_name].add_struct("GENERATE", "data/generate.sh")
            job = self._sim_dict["job"] if box.get_job() is None else box.get_job()
            actuate = Actuate(self._link, box_link, self._sim_dict["cluster"], job, box.get_label(), box.get_struct())
            actuate.generate_files()        

            # Create list for automated filling template
            jinja2_dict = []
            jinja2_dict_fill = []
            area_on = False
            for mol in box.get_mols():
                jinja2_dict.append({"name": mol, "link": "../_gro/"+box.get_struct()[mol].split("/")[-1], "target_dens": str(box.get_mols()[mol][2]), "fill": box.get_mols()[mol][0]=="fill"})
                if box.get_mols()[mol][0]=="fill" and "PORE" in box.get_struct():
                    jinja2_dict_fill.append({"name": mol, "link": "../_gro/"+box.get_struct()[mol].split("/")[-1], "target_dens": str(box.get_mols()[mol][2]), "fill": box.get_mols()[mol][0]=="fill"})
                elif box.get_mols()[mol][0]=="fill" and not box.get_mols()[mol][5]:
                    jinja2_dict_fill.append({"name": mol, "link": "../_gro/"+box.get_struct()[mol].split("/")[-1], "target_dens": str(box.get_mols()[mol][2]), "fill": box.get_mols()[mol][0]=="fill", "box": box.get_mols()[mol][6]})
                elif box.get_mols()[mol][0]=="fill" and box.get_mols()[mol][5]:
                    area_on = True
                    jinja2_dict_fill.append({"name": mol, "link": "../_gro/"+box.get_struct()[mol].split("/")[-1], "target_dens": str(box.get_mols()[mol][2]), "fill": box.get_mols()[mol][0]=="fill", "box": box.get_mols()[mol][6], "area": box.get_mols()[mol][5], "area_length": len(box.get_mols()[mol][5])})


            # Create analysis shell file for automated filling
            if (any(mol["fill"] for mol in jinja2_dict))==True:
                analyze = Analyze(self._link, box_link, self._sim_dict["cluster"])
                analyze.extract_mol("nvt")

                # If a pore system is considered
                if "PORE" in box.get_struct():
                    # Open template with jinja2
                    with open(os.path.split(__file__)[0]+"/templates/auto_dens.py") as file_:
                        template = Template(file_.read())

                    # Adjust template 
                    output = template.render(mols=jinja2_dict, mols2=jinja2_dict_fill, submit=self._sim_dict["cluster"]["queuing"]["submit"]+" min.job", fill = True)
                    #Save adjusted template
                    with open(box_link+"ana/ana.py", "w") as file_:
                        file_.write(output)

                # If a box system is considered     
                else:
                    # Open template with jinja2
                    with open(os.path.split(__file__)[0]+"/templates/auto_dens_box.py") as file_:
                        template = Template(file_.read())

                    # Adjust template 
                    output = template.render(mols=jinja2_dict, mols2=jinja2_dict_fill, submit=self._sim_dict["cluster"]["queuing"]["submit"]+" min.job", fill = True, area = area_on)
                    #Save adjusted template
                    with open(box_link+"ana/ana.py", "w") as file_:
                        file_.write(output)
            
            # If the system should not be filled up
            else:
                if "PORE" in box.get_struct():
                    # Open template with jinja2
                    with open(os.path.split(__file__)[0]+"/templates/auto_dens.py") as file_:
                        template = Template(file_.read())

                    # Adjust template 
                    output = template.render(mols=jinja2_dict, mols2=jinja2_dict_fill, submit=self._sim_dict["cluster"]["queuing"]["submit"]+" min.job", fill =False)
                    #Save adjusted template
                    with open(box_link+"ana/ana.py", "w") as file_:
                        file_.write(output)
                else:
                    analyze = Analyze(self._link, box_link, self._sim_dict["cluster"])
                    analyze.extract_mol("run")
                    with open(os.path.split(__file__)[0]+"/templates/auto_dens_box.py") as file_:
                            template = Template(file_.read())
                    output = template.render(mols=jinja2_dict, submit=self._sim_dict["cluster"]["queuing"]["submit"]+" min.job", fill = False)
                    with open(box_link+"ana/ana.py", "w") as file_:
                            file_.write(output)
                            
            # End message
            print("Finished simulation folder - "+box.get_label()+" ...")


    ##################
    # Setter Methods #
    ##################
    def set_sim_dict(self, sim_dict):
        """Set the simulation dictionary.

        Parameters
        ----------
        sim_dict : dictionary
            Simulation dictionary
        """
        self._sim_dict = sim_dict
        self._update_config()

    def set_box(self, box):
        """Set the simulation box.

        Parameters
        ----------
        box : Box, list
            Simulation box object or a series of those
        """
        boxes = box if isinstance(box, list) else [box]
        self._sim_dict["box"] = {i:boxes[i] for i in range(len(boxes))}
        self._update_config()

    def set_job(self, job):
        """Set the dictionary containing all simulation information.

        Parameters
        ----------
        job : dictionary
            Information dictionary

        Examples
        --------
        Following dictionary structure may be used as an input

        .. code-block:: python

            {"min": {"file": "data/simulation/forhlr.sh", "nodes": 2, "np": 20, "wall": "24:00:00"},
             "nvt": {"file": "data/simulation/forhlr.sh", "nodes": 4, "np": 20, "wall": "24:00:00"},
             "run": {"file": "data/simulation/forhlr.sh", "maxh": 24, "nodes": 11, "np": 20, "runs": 15, "wall": "24:00:00"}}
        """
        self._sim_dict["job"] = job
        self._update_config()

    def set_param(self, param):
        """Set the dictionary containing all systems parameters.

        Parameters
        ----------
        param : dictionary
            Information dictionary

        Examples
        --------
        Following dictionary structure may be used as an input

        .. code-block:: python

            {"min": {"file": "data/simulation/pore_min.mdp"},
             "nvt": {"file": "data/simulation/pore_nvt.mdp", "param": {"NUMBEROFSTEPS": 2000000, "TEMPERATURE_VAL": 298}},
             "run": {"file": "data/simulation/pore_run.mdp", "param": {"NUMBEROFSTEPS": 20000000, "TEMPERATURE_VAL": 298}}}
        """
        self._sim_dict["param"] = param
        self._update_config()

    def set_cluster(self, cluster):
        """Set the dictionary containing all cluster preferences.

        Parameters
        ----------
        cluster : dictionary
            Information dictionary

        Examples
        --------
        Following dictionary structure may be used as an input

        .. code-block:: python

            {"address": "user_name@cluster",
             "directory": "/home/pores/simulation/",
             "queuing": {"add_np": False, "mpi": "$DO_PARALLEL", "shell": "forhlr.sh", "submit": "sbatch --partition multinode"}}
        """
        self._sim_dict["cluster"] = cluster
        self._update_config()


    ##################
    # Getter Methods #
    ##################
    def get_sim_dict(self):
        """Return the simulation dictionary.

        Returns
        -------
        sim_dict : dictionary
            Simulation dictionary
        """
        return self._sim_dict

    def get_box(self):
        """Return the simulation box dictionary.

        Returns
        -------
        box : dictionary
            Simulation box dictionary
        """
        return self._sim_dict["box"]
